@@ -1,5 +1,5 @@
 """
-Open Generation and Transmission Operation and Expansion Planning Model with RES and ESS (openTEPES) - January 19, 2022
+Open Generation and Transmission Operation and Expansion Planning Model with RES and ESS (openTEPES) - February 09, 2022
 """
 
 import time
@@ -128,38 +128,37 @@ def InputData(DirName, CaseName, mTEPES, pIndLogConsole):
     pIndBinGenOperat     = dfOption   ['IndBinGenOperat'    ][0].astype('int')                                                            # Indicator of binary generation operation decisions, 0 continuous - 1 binary
     pIndBinLineCommit    = dfOption   ['IndBinLineCommit'   ][0].astype('int')                                                            # Indicator of binary network    switching decisions, 0 continuous - 1 binary
     pIndNetLosses        = dfOption   ['IndNetLosses'       ][0].astype('int')                                                            # Indicator of network losses,                        0 lossless   - 1 ohmic losses
-    pENSCost             = dfParameter['ENSCost'            ][0] * 1e-3                                                                   # cost of energy not served           [MEUR/GWh]
-    pCO2Cost             = dfParameter['CO2Cost'            ][0]                                                                          # cost of CO2 emission                [EUR/t CO2]
-    pUpReserveActivation = dfParameter['UpReserveActivation'][0]                                                                          # upward   reserve activation         [p.u.]
-    pDwReserveActivation = dfParameter['DwReserveActivation'][0]                                                                          # downward reserve activation         [p.u.]
+    pENSCost             = dfParameter['ENSCost'            ][0] * 1e-3                                                                   # cost of energy not served                [MEUR/GWh]
+    pCO2Cost             = dfParameter['CO2Cost'            ][0]                                                                          # cost of CO2 emission                     [EUR/t CO2]
+    pUpReserveActivation = dfParameter['UpReserveActivation'][0]                                                                          # upward   reserve activation              [p.u.]
+    pDwReserveActivation = dfParameter['DwReserveActivation'][0]                                                                          # downward reserve activation              [p.u.]
     pMinRatioDwUp        = dfParameter['MinRatioDwUp'       ][0]                                                                          # minimum ratio down up operating reserves [p.u.]
     pMaxRatioDwUp        = dfParameter['MaxRatioDwUp'       ][0]                                                                          # maximum ratio down up operating reserves [p.u.]
-    pSBase               = dfParameter['SBase'              ][0] * 1e-3                                                                   # base power                          [GW]
+    pSBase               = dfParameter['SBase'              ][0] * 1e-3                                                                   # base power                               [GW]
     pReferenceNode       = dfParameter['ReferenceNode'      ][0]                                                                          # reference node
-    pTimeStep            = dfParameter['TimeStep'           ][0].astype('int')                                                            # duration of the unit time step      [h]
-    # pStageDuration       = dfParameter['StageDuration'      ][0].astype('int')                                                            # duration of each stage              [h]
+    pTimeStep            = dfParameter['TimeStep'           ][0].astype('int')                                                            # duration of the unit time step           [h]
+    # pStageDuration       = dfParameter['StageDuration'      ][0].astype('int')                                                          # duration of each stage                   [h]
 
-    pScenProb            = dfScenario     ['Probability'   ]                                                                              # probabilities of scenarios          [p.u.]
-    pStageWeight         = dfStage        ['Weight'        ]                                                                              # weights of stages                   [p.u.]
-    pDuration            = dfDuration     ['Duration'      ]    * pTimeStep                                                               # duration of load levels             [h]
-    pReserveMargin       = dfReserveMargin['ReserveMargin' ]                                                                              # minimum adequacy reserve margin     [p.u.]
+    pScenProb            = dfScenario     ['Probability'   ]                                                                              # probabilities of scenarios               [p.u.]
+    pStageWeight         = dfStage        ['Weight'        ]                                                                              # weights of stages                        [p.u.]
+    pDuration            = dfDuration     ['Duration'      ]    * pTimeStep                                                               # duration of load levels                  [h]
+    pReserveMargin       = dfReserveMargin['ReserveMargin' ]                                                                              # minimum adequacy reserve margin          [p.u.]
     pLevelToStage        = dfDuration     ['Stage'         ]                                                                              # load levels assignment to stages
-    pDemand              = dfDemand               [mTEPES.nd]    * 1e-3                                                                   # demand                              [GW]
-    pSystemInertia       = dfInertia              [mTEPES.ar]                                                                             # inertia                             [s]
-    pOperReserveUp       = dfUpOperatingReserve   [mTEPES.ar]    * 1e-3                                                                   # upward   operating reserve          [GW]
-    pOperReserveDw       = dfDwOperatingReserve   [mTEPES.ar]    * 1e-3                                                                   # downward operating reserve          [GW]
-    pVariableMinPower    = dfVariableMinPower     [mTEPES.gg]    * 1e-3                                                                   # dynamic variable minimum power      [GW]
-    pVariableMaxPower    = dfVariableMaxPower     [mTEPES.gg]    * 1e-3                                                                   # dynamic variable maximum power      [GW]
-    pVariableMinCharge   = dfVariableMinCharge    [mTEPES.gg]    * 1e-3                                                                   # dynamic variable minimum charge     [GW]
-    pVariableMaxCharge   = dfVariableMaxCharge    [mTEPES.gg]    * 1e-3                                                                   # dynamic variable maximum charge     [GW]
-    pVariableMinStorage  = dfVariableMinStorage   [mTEPES.gg]                                                                             # dynamic variable minimum storage    [GWh]
-    pVariableMaxStorage  = dfVariableMaxStorage   [mTEPES.gg]                                                                             # dynamic variable maximum storage    [GWh]
-    pEnergyInflows       = dfEnergyInflows        [mTEPES.gg]    * 1e-3                                                                   # dynamic energy inflows              [GW]
-    pEnergyOutflows      = dfEnergyOutflows       [mTEPES.gg]    * 1e-3                                                                   # dynamic energy outflows             [GW]
+    pDemand              = dfDemand               [mTEPES.nd]    * 1e-3                                                                   # demand                                   [GW]
+    pSystemInertia       = dfInertia              [mTEPES.ar]                                                                             # inertia                                  [s]
+    pOperReserveUp       = dfUpOperatingReserve   [mTEPES.ar]    * 1e-3                                                                   # upward   operating reserve               [GW]
+    pOperReserveDw       = dfDwOperatingReserve   [mTEPES.ar]    * 1e-3                                                                   # downward operating reserve               [GW]
+    pVariableMinPower    = dfVariableMinPower     [mTEPES.gg]    * 1e-3                                                                   # dynamic variable minimum power           [GW]
+    pVariableMaxPower    = dfVariableMaxPower     [mTEPES.gg]    * 1e-3                                                                   # dynamic variable maximum power           [GW]
+    pVariableMinCharge   = dfVariableMinCharge    [mTEPES.gg]    * 1e-3                                                                   # dynamic variable minimum charge          [GW]
+    pVariableMaxCharge   = dfVariableMaxCharge    [mTEPES.gg]    * 1e-3                                                                   # dynamic variable maximum charge          [GW]
+    pVariableMinStorage  = dfVariableMinStorage   [mTEPES.gg]                                                                             # dynamic variable minimum storage         [GWh]
+    pVariableMaxStorage  = dfVariableMaxStorage   [mTEPES.gg]                                                                             # dynamic variable maximum storage         [GWh]
+    pEnergyInflows       = dfEnergyInflows        [mTEPES.gg]    * 1e-3                                                                   # dynamic energy inflows                   [GW]
+    pEnergyOutflows      = dfEnergyOutflows       [mTEPES.gg]    * 1e-3                                                                   # dynamic energy outflows                  [GW]
 
     # compute the demand as the mean over the time step load levels and assign it to active load levels. Idem for operating reserve, variable max power, variable min and max storage and inflows
     if pTimeStep > 1:
-        # pDemand.transform(lambda x: np.convolve(x, [.25, 0.25, 0.25, 0.25], 'same'), axis=1)
         pDemand             = pDemand.rolling            (pTimeStep).mean()
         pSystemInertia      = pSystemInertia.rolling     (pTimeStep).mean()
         pOperReserveUp      = pOperReserveUp.rolling     (pTimeStep).mean()
@@ -237,7 +236,7 @@ def InputData(DirName, CaseName, mTEPES, pIndLogConsole):
     pLineType           = dfNetwork     ['LineType'              ]                                                                            # line type
     pLineLength         = dfNetwork     ['Length'                ]                                                                            # line length                                 [km]
     pLineVoltage        = dfNetwork     ['Voltage'               ]                                                                            # line voltage                                [kV]
-    pLineLossFactor     = dfNetwork     ['LossFactor'            ]                                                                            # loss     factor                             [p.u.]
+    pLineLossFactor     = dfNetwork     ['LossFactor'            ]                                                                            # loss factor                                 [p.u.]
     pLineR              = dfNetwork     ['Resistance'            ]                                                                            # resistance                                  [p.u.]
     pLineX              = dfNetwork     ['Reactance'             ].sort_index()                                                               # reactance                                   [p.u.]
     pLineBsh            = dfNetwork     ['Susceptance'           ]                                                                            # susceptance                                 [p.u.]
@@ -279,7 +278,7 @@ def InputData(DirName, CaseName, mTEPES, pIndLogConsole):
     mTEPES.ec = Set(initialize=mTEPES.es,                     ordered=False, doc='candidate   ESS units', filter=lambda mTEPES,es      :  es        in mTEPES.es  and pGenInvestCost    [es] >  0.0)
     mTEPES.br = Set(initialize=mTEPES.ni*mTEPES.nf,           ordered=False, doc='all branches         ', filter=lambda mTEPES,ni,nf   : (ni,nf)    in                pLineType                    )
     mTEPES.ln = Set(initialize=mTEPES.ni*mTEPES.nf*mTEPES.cc, ordered=False, doc='all input       lines', filter=lambda mTEPES,ni,nf,cc: (ni,nf,cc) in                pLineType                    )
-    mTEPES.la = Set(initialize=mTEPES.ln,                     ordered=False, doc='all real        lines', filter=lambda mTEPES,*ln     :  ln        in mTEPES.ln  and pLineX            [ln] >  0.0 and pLineNTCFrw[ln] > 0.0 and pLineNTCBck[ln] > 0.0)
+    mTEPES.la = Set(initialize=mTEPES.ln,                     ordered=False, doc='all real        lines', filter=lambda mTEPES,*ln     :  ln        in mTEPES.ln  and pLineX            [ln] != 0.0 and pLineNTCFrw[ln] > 0.0 and pLineNTCBck[ln] > 0.0)
     mTEPES.lc = Set(initialize=mTEPES.la,                     ordered=False, doc='candidate       lines', filter=lambda mTEPES,*la     :  la        in mTEPES.la  and pNetFixedCost     [la] >  0.0)
     mTEPES.cd = Set(initialize=mTEPES.la,                     ordered=False, doc='             DC lines', filter=lambda mTEPES,*la     :  la        in mTEPES.la  and pNetFixedCost     [la] >  0.0 and pLineType[la] == 'DC')
     mTEPES.ed = Set(initialize=mTEPES.la,                     ordered=False, doc='             DC lines', filter=lambda mTEPES,*la     :  la        in mTEPES.la  and pNetFixedCost     [la] == 0.0 and pLineType[la] == 'DC')
@@ -510,47 +509,54 @@ def InputData(DirName, CaseName, mTEPES, pIndLogConsole):
     pMaxStorage       = pMaxStorage.loc      [mTEPES.sc*mTEPES.p*mTEPES.n          ]
     pIniInventory     = pIniInventory.loc    [mTEPES.sc*mTEPES.p*mTEPES.n          ]
 
+    # separate positive and negative demands to avoid converting negative values to 0
+    pDemandPos          = pDemand.where         (pDemand >= 0.0, other=0.0)
+    pDemandNeg          = pDemand.where         (pDemand <  0.0, other=0.0)
     # small values are converted to 0
-    pPeakDemand = pd.Series(data=[0.0 for ar in mTEPES.ar], index=mTEPES.ar)
-    for a2 in mTEPES.ar:
+    pPeakDemand         = pd.Series(data=[0.0 for ar in mTEPES.ar], index=mTEPES.ar)
+    for ar in mTEPES.ar:
         # values < 1e-5 times the maximum demand for each area (an area is related to operating reserves procurement, i.e., country) are converted to 0
-        pPeakDemand[a2] = pDemand      [[nd for nd in mTEPES.nd if (nd,a2) in mTEPES.ndar]].sum(axis=1).max()
-        pEpsilon        = pPeakDemand[a2]*2.5e-5
+        pPeakDemand[ar] = pDemand      [[nd for nd in mTEPES.nd if (nd,ar) in mTEPES.ndar]].sum(axis=1).max()
+        pEpsilon        = pPeakDemand[ar]*2.5e-5
         # values < 1e-5 times the maximum system demand are converted to 0
         # pEpsilon      = pDemand.sum(axis=1).max()*1e-5
 
         # these parameters are in GW
-        pDemand        [pDemand        [[nd for nd in mTEPES.nd if (nd,a2) in mTEPES.ndar]] < pEpsilon] = 0.0
-        pSystemInertia [pSystemInertia [[                                              a2]] < pEpsilon] = 0.0
-        pOperReserveUp [pOperReserveUp [[                                              a2]] < pEpsilon] = 0.0
-        pOperReserveDw [pOperReserveDw [[                                              a2]] < pEpsilon] = 0.0
-        pMinPower      [pMinPower      [[g  for  g in mTEPES.g  if (a2,g)  in mTEPES.a2g ]] < pEpsilon] = 0.0
-        pMaxPower      [pMaxPower      [[g  for  g in mTEPES.g  if (a2,g)  in mTEPES.a2g ]] < pEpsilon] = 0.0
-        pMinCharge     [pMinCharge     [[es for es in mTEPES.es if (a2,es) in mTEPES.a2g ]] < pEpsilon] = 0.0
-        pMaxCharge     [pMaxCharge     [[es for es in mTEPES.es if (a2,es) in mTEPES.a2g ]] < pEpsilon] = 0.0
-        pEnergyInflows [pEnergyInflows [[es for es in mTEPES.es if (a2,es) in mTEPES.a2g ]] < pEpsilon/pTimeStep] = 0.0
-        pEnergyOutflows[pEnergyOutflows[[es for es in mTEPES.es if (a2,es) in mTEPES.a2g ]] < pEpsilon/pTimeStep] = 0.0
+        pDemandPos     [pDemandPos     [[nd for nd in mTEPES.nd if (nd,ar) in mTEPES.ndar]] <  pEpsilon] = 0.0
+        pDemandNeg     [pDemandNeg     [[nd for nd in mTEPES.nd if (nd,ar) in mTEPES.ndar]] > -pEpsilon] = 0.0
+        pSystemInertia [pSystemInertia [[                                              ar]] <  pEpsilon] = 0.0
+        pOperReserveUp [pOperReserveUp [[                                              ar]] <  pEpsilon] = 0.0
+        pOperReserveDw [pOperReserveDw [[                                              ar]] <  pEpsilon] = 0.0
+        pMinPower      [pMinPower      [[g  for  g in mTEPES.g  if (ar,g)  in mTEPES.a2g ]] <  pEpsilon] = 0.0
+        pMaxPower      [pMaxPower      [[g  for  g in mTEPES.g  if (ar,g)  in mTEPES.a2g ]] <  pEpsilon] = 0.0
+        pMinCharge     [pMinCharge     [[es for es in mTEPES.es if (ar,es) in mTEPES.a2g ]] <  pEpsilon] = 0.0
+        pMaxCharge     [pMaxCharge     [[es for es in mTEPES.es if (ar,es) in mTEPES.a2g ]] <  pEpsilon] = 0.0
+        pEnergyInflows [pEnergyInflows [[es for es in mTEPES.es if (ar,es) in mTEPES.a2g ]] <  pEpsilon/pTimeStep] = 0.0
+        pEnergyOutflows[pEnergyOutflows[[es for es in mTEPES.es if (ar,es) in mTEPES.a2g ]] <  pEpsilon/pTimeStep] = 0.0
+
+        # merging positive and negative values of the demand
+        pDemand         = pDemandPos.where      (pDemandPos >= 0.0, other=pDemandNeg)
 
         # these parameters are in GWh
-        pMinStorage    [pMinStorage    [[es for es in mTEPES.es if (a2,es) in mTEPES.a2g ]] < pEpsilon] = 0.0
-        pMaxStorage    [pMaxStorage    [[es for es in mTEPES.es if (a2,es) in mTEPES.a2g ]] < pEpsilon] = 0.0
-        pIniInventory  [pIniInventory  [[es for es in mTEPES.es if (a2,es) in mTEPES.a2g ]] < pEpsilon] = 0.0
+        pMinStorage    [pMinStorage    [[es for es in mTEPES.es if (ar,es) in mTEPES.a2g ]] <  pEpsilon] = 0.0
+        pMaxStorage    [pMaxStorage    [[es for es in mTEPES.es if (ar,es) in mTEPES.a2g ]] <  pEpsilon] = 0.0
+        pIniInventory  [pIniInventory  [[es for es in mTEPES.es if (ar,es) in mTEPES.a2g ]] <  pEpsilon] = 0.0
 
         # these parameters are in GW
-        for ni,nf,cc,a2 in mTEPES.laar:
+        for ni,nf,cc,ar in mTEPES.laar:
             if  pLineNTCFrw[ni,nf,cc] < pEpsilon:
                 pLineNTCFrw[ni,nf,cc] = 0.0
             if  pLineNTCBck[ni,nf,cc] < pEpsilon:
                 pLineNTCBck[ni,nf,cc] = 0.0
-        for a2,es in mTEPES.a2g:
+        for ar,es in mTEPES.a2g:
             if  pInitialInventory[es] < pEpsilon and es in mTEPES.es:
                 pInitialInventory[es] = 0.0
 
         pMaxPower2ndBlock  = pMaxPower  - pMinPower
         pMaxCharge2ndBlock = pMaxCharge - pMinCharge
 
-        pMaxPower2ndBlock [pMaxPower2ndBlock [[es for es in mTEPES.es if (a2,es) in mTEPES.a2g]] < pEpsilon] = 0.0
-        pMaxCharge2ndBlock[pMaxCharge2ndBlock[[es for es in mTEPES.es if (a2,es) in mTEPES.a2g]] < pEpsilon] = 0.0
+        pMaxPower2ndBlock [pMaxPower2ndBlock [[es for es in mTEPES.es if (ar,es) in mTEPES.a2g]] < pEpsilon] = 0.0
+        pMaxCharge2ndBlock[pMaxCharge2ndBlock[[es for es in mTEPES.es if (ar,es) in mTEPES.a2g]] < pEpsilon] = 0.0
 
     # replace < 0.0 by 0.0
     pMaxPower2ndBlock  = pMaxPower2ndBlock.where (pMaxPower2ndBlock  > 0.0, other=0.0)
@@ -589,18 +595,18 @@ def InputData(DirName, CaseName, mTEPES, pIndLogConsole):
 
     mTEPES.pENSCost              = Param(initialize=pENSCost             , within=NonNegativeReals)
     mTEPES.pCO2Cost              = Param(initialize=pCO2Cost             , within=NonNegativeReals)
-    mTEPES.pUpReserveActivation  = Param(initialize=pUpReserveActivation , within=NonNegativeReals)
-    mTEPES.pDwReserveActivation  = Param(initialize=pDwReserveActivation , within=NonNegativeReals)
-    mTEPES.pMinRatioDwUp         = Param(initialize=pMinRatioDwUp        , within=NonNegativeReals)
-    mTEPES.pMaxRatioDwUp         = Param(initialize=pMaxRatioDwUp        , within=NonNegativeReals)
+    mTEPES.pUpReserveActivation  = Param(initialize=pUpReserveActivation , within=UnitInterval    )
+    mTEPES.pDwReserveActivation  = Param(initialize=pDwReserveActivation , within=UnitInterval    )
+    mTEPES.pMinRatioDwUp         = Param(initialize=pMinRatioDwUp        , within=UnitInterval    )
+    mTEPES.pMaxRatioDwUp         = Param(initialize=pMaxRatioDwUp        , within=UnitInterval    )
     mTEPES.pSBase                = Param(initialize=pSBase               , within=NonNegativeReals)
     mTEPES.pTimeStep             = Param(initialize=pTimeStep            , within=NonNegativeReals)
     # mTEPES.pStageDuration        = Param(initialize=pStageDuration      , within=NonNegativeReals)
 
     mTEPES.pReserveMargin        = Param(                               mTEPES.ar, initialize=pReserveMargin.to_dict()            , within=NonNegativeReals, doc='Adequacy reserve margin'        )
     mTEPES.pPeakDemand           = Param(                               mTEPES.ar, initialize=pPeakDemand.to_dict()               , within=NonNegativeReals, doc='Peak demand'                    )
-    mTEPES.pDemand               = Param(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.nd, initialize=pDemand.stack().to_dict()           , within=NonNegativeReals, doc='Demand'                         )
-    mTEPES.pScenProb             = Param(mTEPES.scc,                               initialize=pScenProb.to_dict()                 , within=NonNegativeReals, doc='Probability'                    )
+    mTEPES.pDemand               = Param(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.nd, initialize=pDemand.stack().to_dict()           , within=           Reals, doc='Demand'                         )
+    mTEPES.pScenProb             = Param(mTEPES.scc,                               initialize=pScenProb.to_dict()                 , within=UnitInterval    , doc='Probability'                    )
     mTEPES.pStageWeight          = Param(mTEPES.stt,                               initialize=pStageWeight.to_dict()              , within=NonNegativeReals, doc='Stage weight'                   )
     mTEPES.pDuration             = Param(                     mTEPES.n,            initialize=pDuration.to_dict()                 , within=NonNegativeReals, doc='Duration'                       )
     mTEPES.pNodeLon              = Param(                               mTEPES.nd, initialize=pNodeLon.to_dict()                  ,                          doc='Longitude'                      )
@@ -621,8 +627,8 @@ def InputData(DirName, CaseName, mTEPES, pIndLogConsole):
     mTEPES.pRatedMaxPower        = Param(                               mTEPES.gg, initialize=pRatedMaxPower.to_dict()            , within=NonNegativeReals, doc='Rated maximum power'            )
     mTEPES.pMustRun              = Param(                               mTEPES.gg, initialize=pMustRun.to_dict()                  , within=Boolean         , doc='must-run unit'                  )
     mTEPES.pInertia              = Param(                               mTEPES.gg, initialize=pInertia.to_dict()                  , within=NonNegativeReals, doc='unit inertia constant'          )
-    mTEPES.pAvailability         = Param(                               mTEPES.gg, initialize=pAvailability.to_dict()             , within=NonNegativeReals, doc='unit availability',           mutable=True)
-    mTEPES.pEFOR                 = Param(                               mTEPES.gg, initialize=pEFOR.to_dict()                     , within=NonNegativeReals, doc='EFOR'                           )
+    mTEPES.pAvailability         = Param(                               mTEPES.gg, initialize=pAvailability.to_dict()             , within=UnitInterval    , doc='unit availability',           mutable=True)
+    mTEPES.pEFOR                 = Param(                               mTEPES.gg, initialize=pEFOR.to_dict()                     , within=UnitInterval    , doc='EFOR'                           )
     mTEPES.pLinearOperCost       = Param(                               mTEPES.gg, initialize=pLinearOperCost.to_dict()           , within=NonNegativeReals, doc='Linear   variable cost'         )
     mTEPES.pLinearVarCost        = Param(                               mTEPES.gg, initialize=pLinearVarCost.to_dict()            , within=NonNegativeReals, doc='Linear   variable cost'         )
     mTEPES.pLinearOMCost         = Param(                               mTEPES.gg, initialize=pLinearOMCost.to_dict()             , within=NonNegativeReals, doc='Linear   O&M      cost'         )
@@ -642,7 +648,7 @@ def InputData(DirName, CaseName, mTEPES, pIndLogConsole):
     mTEPES.pIndBinUnitRetire     = Param(                               mTEPES.gg, initialize=pIndBinUnitRetire.to_dict()         , within=Boolean         , doc='Binary retirement decision'     )
     mTEPES.pIndBinUnitCommit     = Param(                               mTEPES.gg, initialize=pIndBinUnitCommit.to_dict()         , within=Boolean         , doc='Binary commitment decision'     )
     mTEPES.pIndOperReserve       = Param(                               mTEPES.gg, initialize=pIndOperReserve.to_dict()           , within=Boolean         , doc='Indicator of operating reserve' )
-    mTEPES.pEfficiency           = Param(                               mTEPES.gg, initialize=pEfficiency.to_dict()               , within=NonNegativeReals, doc='Round-trip efficiency'          )
+    mTEPES.pEfficiency           = Param(                               mTEPES.gg, initialize=pEfficiency.to_dict()               , within=UnitInterval    , doc='Round-trip efficiency'          )
     mTEPES.pCycleTimeStep        = Param(                               mTEPES.gg, initialize=pCycleTimeStep.to_dict()            , within=NonNegativeReals, doc='ESS Storage cycle'              )
     mTEPES.pOutflowsTimeStep     = Param(                               mTEPES.gg, initialize=pOutflowsTimeStep.to_dict()         , within=NonNegativeReals, doc='ESS Outflows cycle'             )
     mTEPES.pIniInventory         = Param(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.gg, initialize=pIniInventory.stack().to_dict()     , within=NonNegativeReals, doc='ESS Initial storage',         mutable=True)
@@ -651,7 +657,7 @@ def InputData(DirName, CaseName, mTEPES, pIndLogConsole):
 
     mTEPES.pLineLossFactor       = Param(                               mTEPES.ln, initialize=pLineLossFactor.to_dict()           , within=NonNegativeReals, doc='Loss factor'                    )
     mTEPES.pLineR                = Param(                               mTEPES.ln, initialize=pLineR.to_dict()                    , within=NonNegativeReals, doc='Resistance'                     )
-    mTEPES.pLineX                = Param(                               mTEPES.ln, initialize=pLineX.to_dict()                    , within=NonNegativeReals, doc='Reactance'                      )
+    mTEPES.pLineX                = Param(                               mTEPES.ln, initialize=pLineX.to_dict()                    , within=           Reals, doc='Reactance'                      )
     mTEPES.pLineBsh              = Param(                               mTEPES.ln, initialize=pLineBsh.to_dict()                  , within=NonNegativeReals, doc='Susceptance',                 mutable=True)
     mTEPES.pLineTAP              = Param(                               mTEPES.ln, initialize=pLineTAP.to_dict()                  , within=NonNegativeReals, doc='Tap changer',                 mutable=True)
     mTEPES.pLineLength           = Param(                               mTEPES.ln, initialize=pLineLength.to_dict()               , within=NonNegativeReals, doc='Length',                      mutable=True)
@@ -713,7 +719,7 @@ def SettingUpVariables(OptModel, mTEPES):
     OptModel.vCharge2ndBlock       = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.es, within=NonNegativeReals, bounds=lambda OptModel,sc,p,n,es: (0.0,mTEPES.pMaxCharge2ndBlock[sc,p,n,es]),                    doc='ESS       charge power                           [GW]')
     OptModel.vESSReserveUp         = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.es, within=NonNegativeReals, bounds=lambda OptModel,sc,p,n,es: (0.0,mTEPES.pMaxCharge2ndBlock[sc,p,n,es]),                    doc='ESS upward   operating reserve                   [GW]')
     OptModel.vESSReserveDown       = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.es, within=NonNegativeReals, bounds=lambda OptModel,sc,p,n,es: (0.0,mTEPES.pMaxCharge2ndBlock[sc,p,n,es]),                    doc='ESS downward operating reserve                   [GW]')
-    OptModel.vENS                  = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.nd, within=NonNegativeReals, bounds=lambda OptModel,sc,p,n,nd: (0.0,mTEPES.pDemand           [sc,p,n,nd]),                    doc='energy not served in node                        [GW]')
+    OptModel.vENS                  = Var(mTEPES.sc, mTEPES.p, mTEPES.n, mTEPES.nd, within=NonNegativeReals, bounds=lambda OptModel,sc,p,n,nd: (0.0,max(0.0,mTEPES.pDemand   [sc,p,n,nd])),                   doc='energy not served in node                        [GW]')
 
     if mTEPES.pIndBinGenInvest() == 0:
         OptModel.vGenerationInvest = Var(                               mTEPES.gc, within=UnitInterval,                                                                                                      doc='generation      investment decision exists in a year [0,1]')
@@ -806,16 +812,22 @@ def SettingUpVariables(OptModel, mTEPES):
             OptModel.vOutput2ndBlock[sc,p,n,nr].fix(0.0)
             OptModel.vReserveUp     [sc,p,n,nr].fix(0.0)
             OptModel.vReserveDown   [sc,p,n,nr].fix(0.0)
+        if  mTEPES.pIndOperReserve  [       nr] !=  0.0:
+            OptModel.vReserveUp     [sc,p,n,nr].fix(0.0)
+            OptModel.vReserveDown   [sc,p,n,nr].fix(0.0)
 
     for sc,p,n,es in mTEPES.sc*mTEPES.p*mTEPES.n*mTEPES.es:
         # ESS with no charge capacity or not storage capacity can't charge
-        if mTEPES.pMaxCharge[sc,p,n,es] == 0.0:
+        if mTEPES.pMaxCharge        [sc,p,n,es] == 0.0:
             OptModel.vESSTotalCharge[sc,p,n,es].fix(0.0)
         if mTEPES.pMaxCharge2ndBlock[sc,p,n,es] == 0.0:
             OptModel.vCharge2ndBlock[sc,p,n,es].fix(0.0)
             OptModel.vESSReserveUp  [sc,p,n,es].fix(0.0)
             OptModel.vESSReserveDown[sc,p,n,es].fix(0.0)
-        if mTEPES.pMaxStorage[sc,p,n,es] == 0.0:
+        if  mTEPES.pIndOperReserve  [       es] !=  0.0:
+            OptModel.vESSReserveUp  [sc,p,n,es].fix(0.0)
+            OptModel.vESSReserveDown[sc,p,n,es].fix(0.0)
+        if mTEPES.pMaxStorage       [sc,p,n,es] == 0.0:
             OptModel.vESSInventory  [sc,p,n,es].fix(0.0)
 
     # thermal and RES units ordered by increasing variable operation cost, excluding reactive generating units
@@ -916,7 +928,7 @@ def SettingUpVariables(OptModel, mTEPES):
 
     # fixing the ENS in nodes with no demand
     for sc,p,n,nd in mTEPES.sc*mTEPES.p*mTEPES.n*mTEPES.nd:
-        if mTEPES.pDemand[sc,p,n,nd] ==  0.0:
+        if mTEPES.pDemand[sc,p,n,nd] == 0.0:
             OptModel.vENS  [sc,p,n,nd].fix(0.0)
 
     SettingUpVariablesTime = time.time() - StartTime
